%SUPERRESRECONSTRUCT   Reconstruct a super resolution image from a PALM/STORM sequence
%
% SYNOPSIS:
%  [coords,lowres,highresHist1,highresHist2] = superresreconstruct_ics(in,bgin,gain,PSF_FWHM,Photons,zm)
%
% PARAMETERS:
%  in               : input file name of the data (*.ics)
%  bgin             : Background image or value
%  gain             : CCD gain [counts/photon]
%  PSF_FWHM         : PSF FWHM [px]
%                   (used for the automated selection of single emitter regions)
%  Photons          : Expected Photon count of a single emitter 
%                   (used for the automated selection of single emitter regions)
%  zoom             : Up sample factor for the high resolution image [default 10]
%
% OUTPUT:
%  coords           : Localizations, [x y t I sigma bg sigma_psf idiv]
%  lowres           : Sum of raw data frames
%  highresHist1     : Superresolution image of binned localizations
%  highresHist2     : Superresolution image of binned localizations with histogram equalization
%
% Bernd Rieger,Keith Lidke Jan 2010

function varargout = superresreconstruct_ics(varargin)

d = struct('menu','SingleMolecule',...
    'display','Reconstruct Superresolution Image',...
    'inparams',struct('name',       {'fname_in','bgin','gain','PSF_FWHM','Photons','zm'},...
    'description',{'Input image','Background image/value','Gain [counts/photon]','PSF FWHM [px]','Minimum Photon count', 'High resolution zoom'},...
    'type',       {'string','image','array','array','array','array'},...
    'dim_check',  {0,0,0,0,0,0},...
    'range_check',{'*.ics',[],'R+','R+','R+','N+'},...
    'required',   {1,1,1,1,0,0},...
    'default',    {'','',1,1,100,10}...
    ),...
    'outparams',struct('name',{'coords','lowres','highresHist1','highresHist2'},...
    'description',{'Coordinates','Low res Image','High res Image (Smoothed Histogram)','High res Image (Equalized Histogram)'},...
    'type',{'array','image','image','image'}...
    )...
    );

if nargin == 1
    s = varargin{1};
    if ischar(s) & strcmp(s,'DIP_GetParamList')
        varargout{1} = d;
        return
    end
end
try
    [fname_in,bgin,gain,PSF_FWHM,Photons,zm] = getparams(d,varargin{:});
catch
    if ~isempty(paramerror)
        error(paramerror)
    else
        error(firsterr)
    end
end

PSFsigma = PSF_FWHM/(2*sqrt(2*log(2)));

% ---- some hard coded magic parameters --
fittype = 2;
MIN_PHOTONS=100;        % ignore emitters with intensities<MIN_PHOTONS
FAC_SIGMA = 2;          % ignore emitters with sigmas<PSFsigma/FAC_SIGMA;
MARGIN_WIDTH = 0.5;     % ignore fits with |model PSF width/PSFsigma-1|>MARGIN_WIDTH
% MARGIN_WIDTH = 1.0;     % ignore fits with |model PSF width/PSFsigma-1|>MARGIN_WIDTH
MAX_BG = 500;
MIN_PVALUE = 0.01;

ppt=4; %pixel per thread on the GPU, datasize must divide without remainder by 4
NtimeBlock = 1500;
Pixels = floor((6*PSFsigma+1)/2)*2+1;
minI = Photons/2/3/3/PSFsigma/PSFsigma;

switch fittype
    case 1;     k=Pixels^2-4;
    case {2,3}; k=Pixels^2-5;
    case 4;     k=Pixels^2-6;
end
MIN_IDIV = -0.5*chi2inv(1-MIN_PVALUE,k);

% ----------------------------------------
file_props =  dipio_imagefilegetinfo(fname_in,'ics',0);
Nframes = file_props.size(3);

if isempty(imsize(bgin))
    bg = single(bgin);
else
    bgin = single(bgin);
    bg = repmat(bgin,[1,1,szin(3)]);
end

Ntotal=0;
t_offset = 0;

Xt=[];Yt=[];T=[];sigCollect=[];intCollect=[];bgCollect=[];idivCollect=[];
if fittype == 2
    S_allCollect=[];
end

lowres = cell(ceil(Nframes/NtimeBlock),1);

% Iterate over parts of the dataset
for ii=1:ceil(Nframes/NtimeBlock)
    fprintf(' -- Processing %d of %d --\n',ii,ceil(Nframes/NtimeBlock));
       
    % Read data
    stack = dipio_imagereadroi(fname_in,[0 0 t_offset],[file_props.size(1:2) NtimeBlock],[1 1 1],'ics',0);
       
    % ---- recut to mod ppt ----------
    szin = size(stack);
    sz = szin(1:2) - mod(szin(1:2),ppt);
    data = single((stack(0:sz(1)-1, 0:sz(2)-1,:))-bg)/gain;
    sz = size(data);
    
    allcds=[];    
    
    tic %segmentation
    cds = FindCandidates_v2(data,PSFsigma,Photons,Pixels);
    t_Seg(ii)=toc;
    
    tic%parsing subregions
    [in_ims l t]=cMakeSubregions(cds(:,1),cds(:,2),cds(:,3),Pixels,data);
    t_Parse(ii)=toc;
    
    %tic%CPU fitting
    %[ycpu xcpu]=CgaussMLE(in,PSFsigma,10);
    %t_CPUfit(ii)=toc
    
    tic%GPU fitting
    [P CRLB idiv] = mex_GPUgaussMLEv2(in_ims,PSFsigma,10,fittype);
    
    t_GPUfit(ii)=toc;
        
    x = P(:,1);
    y = P(:,2);
    N_all = P(:,3);
    bg_all = P(:,4);
    
    lay = CRLB(:,1);
    lax = CRLB(:,2);
    li = CRLB(:,3);
    lb = CRLB(:,4);
    
    Ntotal = Ntotal + length(y);
    sigmas = sqrt(lax+lay);
    
    % Filter localizations    
    mask = N_all > MIN_PHOTONS;
    mask = (sigmas<PSFsigma/FAC_SIGMA) & mask;
    mask = (sigmas>0) & mask;
    mask = (bg_all<MAX_BG) & mask;
%     mask = (idiv>MIN_IDIV) & mask;
%     
    if fittype == 2
        mask = mask & (P(:,5) < ((1+MARGIN_WIDTH)*PSFsigma));
        mask = mask & (P(:,5) > ((1-MARGIN_WIDTH)*PSFsigma));
        S_allCollect = cat(1,S_allCollect,P(mask,5));
    end
    
    %fprintf('  size sigmas %d %d, x %d %d\n',size(sigmas),size(x))
    %fprintf('  mask %d\n',sum(mask))
    
    sigCollect = cat(1, sigCollect, sigmas(mask));
    intCollect = cat(1, intCollect, N_all(mask));
    bgCollect = cat(1 ,bgCollect, bg_all(mask));
    idivCollect = cat(1, idivCollect, idiv(mask));
    
    X=x(mask)+l(mask);
    Y=y(mask)+t(mask);
    
    Xt = cat(1,Xt,X); 
    Yt = cat(1,Yt,Y);
    T = cat(1,T,cds(mask,3)+t_offset);
    lowres{ii} = squeeze(sum(mat2im(data),[],[3]));
    
    % Find out the starting time of the next sequence
    t_offset = t_offset + NtimeBlock;
end

Xt = Xt+0.5;
Yt = Yt+0.5;
T = T + 1;

switch fittype
    case 1
        varargout{1} = cat(2,Yt,Xt,T,intCollect,sigCollect,bgCollect,idivCollect);
    case 2    
        varargout{1} = cat(2,Yt,Xt,T,intCollect,sigCollect,bgCollect,S_allCollect,idivCollect);
    otherwise
        printf('No outputs possible because fittype was not chosen properly');
end

totalTime = sum(t_GPUfit)+ sum(t_Parse)+ sum(t_Seg);
fprintf('total time: GPU %2.2f, Parse %2.2f, Seg %2.2f\n',sum(t_GPUfit), sum(t_Parse), sum(t_Seg));
fprintf('percentage: GPU %2.2f, Parse %2.2f, Seg %2.2f\n', [sum(t_GPUfit), sum(t_Parse), sum(t_Seg)]./totalTime*100)
fprintf('fits pers second %e\n',Ntotal/sum(t_GPUfit))
fprintf('Number of fits %d, number of total accepted single events %d (ratio %2.2f)\n',Ntotal, length(Xt),Ntotal/length(Xt))

[o1]=cHistRecon(sz(1)*zm,sz(2)*zm,single(Xt*zm),single(Yt*zm),1);
highresImage1 = mat2im(o1);

[o1,o2]=cHistRecon(sz(1)*zm,sz(2)*zm,single(Xt*zm),single(Yt*zm),0);
highresImage2 = gaussf(mat2im(o2), double(median(sigCollect)*zm));

lowresImage = lowres{1};
for ii=2:ceil(Nframes/NtimeBlock)
    lowresImage = lowresImage+lowres{ii};
end
varargout{2} =  lowresImage;
if nargout>2
    varargout{3} = highresImage1;
end
if nargout>3
    varargout{4} = highresImage2;
end


